import io
import swiftclient
import swiftclient.service
import wandio.file

CHUNK_SIZE = 1 * 1024 * 1024

SEGMENT_SIZE = 1073741824
SEGMENT_CONTAINER_TMPL = ".%s-segments"

DEFAULT_OPTIONS = {
    "os_auth_url": "https://hermes-auth.caida.org",
    "auth_version": "3",
}


def process_options(options=None):
    # following is borrowed from swiftclient.service.SwiftService
    default_opts = dict(
        swiftclient.service._default_global_options,
        **dict(
            swiftclient.service._default_local_options,
            **DEFAULT_OPTIONS
        )
    )
    if options is not None:
        options = dict(
            default_opts,
            **options
        )
    else:
        options = default_opts
    swiftclient.service.process_options(options)

    return options


def get_auth(options=None, connection=None):
    """
    Get the auth URL and auth token for the given options or connection
    :param options:
    :param connection:
    :return:
    """
    if connection is None:
        connection = get_connection(options)
    return connection.get_auth()


def get_service(options=None):
    """
    Get a ready-to-use SwiftService instance
    :param options:
    :return:
    """
    options = process_options(options)
    return swiftclient.service.SwiftService(options=options)


def get_connection(options=None):
    """
    Get a swift Connection instance
    :param options:
    :return:
    """
    options = process_options(options)
    return swiftclient.service.get_conn(options)


def parse_url(url):
    """
    Parse a 'swift://CONTAINER/OBJECT' style URL
    :param url:
    :return: dictionary with "container" and "obj" keys
    """
    url = url.replace("swift://", "")
    if url.find("/") == -1:
        raise ValueError("Swift url must be 'swift://container/object'")
    pieces = url.split("/")
    containername = pieces[0]
    objname = "/".join(pieces[1:])
    return {
        "container": containername,
        "obj": objname,
    }


def list(container=None, options=None, swift=None):
    """
    Get a list of objects in the account or container
    :param container: container to list (if None, the account will be listed)
    :param options:
    :param swift:
    :return:
    """
    if swift is None:
        swift = get_service(options)
    for page in swift.list(container=container):
        if page["success"]:
            for item in page["listing"]:
                yield item["name"]
        else:
            raise page["error"]


def stat(container=None, objects=None, options=None, swift=None):
    """
    Get stats for a list of objects in the account or container
    :param container: container to stat (if None, the account will be listed)
    :param objects: objects to get stats for (if None, the container will be listed)
    :param options:
    :param swift:
    :return:
    """
    if swift is None:
        swift = get_service(options)
    stat_res = swift.stat(container=container, objects=objects)
    if objects is None:
        if not stat_res["success"]:
            raise stat_res["error"]
        yield stat_res
    else:
        for obj_stat in stat_res:
            if not obj_stat["success"]:
                raise obj_stat["error"]
            yield obj_stat


def upload(local_file, container, obj, options=None, swift=None):
    """
    Upload a local file (or file-like object) to the given container and object
    :param local_file: path to local file, or file-like object to upload
    :param container: container to upload to
    :param obj: object name to upload to
    :param options: swift options
    :param swift: existing swift service instance to use
    :return:
    """
    if swift is None:
        swift = get_service(options)
    suo = swiftclient.service.SwiftUploadObject(local_file, object_name=obj)
    results = swift.upload(container, [suo], options={
        "segment_size": SEGMENT_SIZE,
        "segment_container": SEGMENT_CONTAINER_TMPL % container,
    })
    for res in results:
        if not res["success"]:
            # Failing to create a container is a warning, not an error -- Shane
            # Ref: https://github.com/openstack/python-swiftclient/blob/master/swiftclient/shell.py
            # inside function st_upload()
            if 'action' in res and res['action'] == "create_container":
                continue
            else:
                raise res["error"]


def download(container, obj, local_file=None, local_dir=None,
             remove_prefix=True, options=None, swift=None):
    """
    Download an object from swift to the given local file or directory
    :param container: container to download from
    :param obj: object name to download
    :param local_file: path to local file to download to
    :param local_dir: path to local directory to download to
    :param remove_prefix: if local_dir is set, remove the prefix from the object name
    :param options: swift options
    :param swift: existing swift service instance to use
    :return:
    """
    if swift is None:
        swift = get_service(options)
    if local_file is not None:
        opts = {"out_file": local_file}
    elif local_dir is not None:
        opts = {"out_directory": local_dir, "remove_prefix": remove_prefix}
    else:
        raise ValueError("Either localfile or localdir argument must be given")
    results = swift.download(container, [obj], options=opts)
    for res in results:
        if not res["success"]:
            raise res["error"]


def delete(container=None, objects=None, options=None, swift=None):
    """
    Delete a container or a list of objects in the given container
    :param container: container to delete (from)
    :param objects: objects to delete (if None, the container will be deleted)
    :param options:
    :param swift:
    :return:
    """
    if swift is None:
        swift = get_service(options)
    res_iter = swift.delete(container=container, objects=objects)
    for res in res_iter:
        if not res["success"]:
            raise res["error"]


class SwiftReader(wandio.file.GenericReader):

    def __init__(self, url, options=None):
        self.conn = get_connection(options)
        (hdr, body) = self.conn.get_object(resp_chunk_size=CHUNK_SIZE,
                                           **parse_url(url))
        super(SwiftReader, self).__init__(body)

    def close(self):
        if self.conn:
            self.conn.close()


# TODO: figure out how to stream to swift rather than buffer all in memory
class SwiftWriter(wandio.file.GenericWriter):

    def __init__(self, url, options=None, use_bytes_io=False):
        parsed_url = parse_url(url)
        self.container = parsed_url["container"]
        self.object = parsed_url["obj"]
        self.options = options
        if use_bytes_io:
            self.buffer = io.BytesIO()
        else:
            self.buffer = io.StringIO()
        super(SwiftWriter, self).__init__(self.buffer)

    def flush(self):
        pass

    def close(self):
        self.buffer.seek(0)
        upload(self.buffer, container=self.container,
               obj=self.object, options=self.options)



import wandio

if __name__ == '__main__':

    files = [
        'swift://datasets-external-netacq-codes/country_codes.csv',
        'https://www.caida.org/~mingwei/',
        'http://data.caida.org/datasets/as-relationships/README.txt',
        'http://loki.caida.org:2243/data/external/as-rank-ribs/19980101/19980101.as-rel.txt.bz2'
    ]

    for filename in files:
        # the with statement automatically closes the file at the end
        # of the block
        try:
            with wandio.open(filename) as fh:
                line_count = 0
                word_count = 0
                for line in fh:
                    word_count += len(line.rstrip().split())
                    line_count +=1
            # print the number of lines and words in file
            print(filename)
            print(line_count, word_count)
        except IOError as err:
            print(filename)
            raise err



import wandio

if __name__ == '__main__':

    with wandio.open('http://data.caida.org/datasets/as-relationships/README.txt') as fh:
        with wandio.open('test.txt.gz', mode='w') as ofh:
            line_count = 0
            word_count = 0
            for line in fh:
                word_count += len(line.rstrip().split())
                line_count +=1
                ofh.write(line)
            print(line_count, word_count)



import os
import stat
import sys


def file_stat(filename):
    sr = os.stat(filename)
    return {
        "size": sr[stat.ST_SIZE],
        "mtime": sr[stat.ST_MTIME],
    }


class GenericReader(object):
    """
    Wraps a file-like object
    """

    def __init__(self, fh):
        self.fh = fh

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.fh)

    def next(self):
        return self.__next__()

    def read(self, *args):
        return self.fh.read(*args)

    def readline(self):
        return self.fh.readline()

    def close(self):
        self.fh.close()


class GenericWriter(object):
    """
    Wraps a file-like writer object
    """

    def __init__(self, fh):
        self.fh = fh

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def flush(self):
        self.fh.flush()

    def write(self, data):
        self.fh.write(data)

    def writelines(self, lines):
        self.fh.writelines(lines)

    def close(self):
        self.fh.close()


class StdinReader(GenericReader):

    def __init__(self):
        super(StdinReader, self).__init__(sys.stdin)


class SimpleReader(GenericReader):

    def __init__(self, filename):
        super(SimpleReader, self).__init__(open(filename, "r"))


class SimpleWriter(GenericWriter):

    def __init__(self, filename):
        super(SimpleWriter, self).__init__(open(filename, "w"))



#!/usr/bin/env python

import argparse
import os, sys, errno
import shutil
import socket

# urllib import compatible with both python2 and python3
# https://python-future.org/compatible_idioms.html#urllib-module
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

import wandio.compressed
import wandio.file
import wandio.http
import wandio.swift


class Reader(wandio.file.GenericReader):

    def __init__(self, filename, options=None):
        self.filename = filename

        # check for the transport types first (HTTP, Swift, Simple)

        # is this Swift
        if filename.startswith("swift://"):
            fh = wandio.swift.SwiftReader(self.filename, options=options)

        # is this simple HTTP ?
        elif urlparse(self.filename).netloc:
            fh = wandio.http.HttpReader(self.filename)

        # stdin?
        elif filename == "-":
            fh = wandio.file.StdinReader()

        # then it must be a simple local file
        else:
            fh = wandio.file.SimpleReader(self.filename)

        assert fh

        # now check the encoding types (gzip, bzip, plain)

        # Gzip?
        if filename.endswith(".gz"):
            fh = wandio.compressed.GzipReader(fh)

        # Bzip2?
        elif filename.endswith(".bz2"):
            fh = wandio.compressed.BzipReader(fh)

        # Plain, leave the transport handle as-is
        else:
            pass

        super(Reader, self).__init__(fh)


# TODO: refactor Reader and Writer
class Writer(wandio.file.GenericWriter):

    def __init__(self, filename, options=None):
        self.filename = filename

        # check for the transport types first (HTTP, Swift, Simple)

        # is this Swift
        if filename.startswith("swift://"):
            use_bytes_io = False
            if filename.endswith(".gz") or filename.endswith(".bz2"):
                use_bytes_io = True
            fh = wandio.swift.SwiftWriter(self.filename, options=options, use_bytes_io=use_bytes_io)

        # is this simple HTTP ?
        elif urlparse(self.filename).netloc:
            raise NotImplementedError("Writing to HTTP is not supported")

        # then it must be a simple local file
        else:
            fh = wandio.file.SimpleWriter(self.filename)

        assert fh

        # now check the encoding types (gzip, bzip, plain)

        # Gzip?
        if filename.endswith(".gz"):
            fh = wandio.compressed.GzipWriter(fh)

        # Bzip2?
        elif filename.endswith(".bz2"):
            fh = wandio.compressed.BzipWriter(fh)

        # Plain, leave the transport handle as-is
        else:
            pass

        super(Writer, self).__init__(fh)


def wandio_open(filename, mode="r"):
    if mode == "r":
        return Reader(filename)
    elif mode == "w":
        return Writer(filename)
    else:
        raise ValueError("Invalid mode. Mode must be either 'r' or 'w'")


def wandio_stat(filename):
    # currently we support a *very* limited set stat fields:
    # - mtime (Last-Modified for HTTP)
    # - size
    # is this Swift
    if filename.startswith("swift://"):
        # TODO
        raise NotImplementedError("Stat not yet supported for Swift files")

    # is this simple HTTP ?
    elif urlparse(filename).netloc:
        statfunc = wandio.http.http_stat

    # stdin?
    elif filename == "-":
        raise NotImplementedError("Cannot perform stat operation on STDIN")

    # then it must be a simple local file
    else:
        statfunc = wandio.file.file_stat

    return statfunc(filename)


def read_main():
    parser = argparse.ArgumentParser(description="""
    Reads from a file (or files) and writes its contents to stdout. Supports
    any compression/transport that pywandio supports. E.g. HTTP, Swift, gzip, bzip
    """)

    parser.add_argument('-l', '--use-readline', required=False,
                        action='store_true',
                        help="Force use of readline (for testing)")

    parser.add_argument('-n', '--use-next', required=False,
                        action='store_true',
                        help="Force use of next (for testing)")

    parser.add_argument('files', nargs='+', help='Files to read from')

    opts = vars(parser.parse_args())

    for filename in opts['files']:
        with Reader(filename) as fh:
            if opts['use_next']:
                # sys.stderr.write("Reading using 'next'\n")
                for line in fh:
                    sys.stdout.write(line)
            elif opts['use_readline']:
                # sys.stderr.write("Reading using 'readline'\n")
                line = fh.readline()
                while line:
                    sys.stdout.write(line)
                    line = fh.readline()
            else:
                # sys.stderr.write("Reading using 'shutil'\n")
                if (sys.version_info > (3, 0)):
                    try:
                        shutil.copyfileobj(fh, sys.stdout.buffer)
                    except BrokenPipeError:
                        devnull = os.open(os.devnull, os.O_WRONLY)
                        os.dup2(devnull, sys.stdout.fileno())
                        sys.exit(1)
                else:
                    try:
                        shutil.copyfileobj(fh, sys.stdout)
                    except IOError as e:
                        if e.errno != errno.EPIPE:
                            # if it is not a BrokenPipeError, raise the error.
                            raise e
                        devnull = os.open(os.devnull, os.O_WRONLY)
                        os.dup2(devnull, sys.stdout.fileno())
                        sys.exit(1)

def write_main():
    parser = argparse.ArgumentParser(description="""
    Reads from stdin and writes to a file. Supports any compression/transport
    that pywandio supports. E.g. HTTP, Swift, gzip, bzip
    """)

    parser.add_argument('file', help='File to write to')

    opts = vars(parser.parse_args())

    with Writer(opts["file"]) as out_fh:
        with Reader("-") as in_fh:
            for line in in_fh:
                out_fh.write(line)


def stat_main():
    parser = argparse.ArgumentParser(description="""
    Obtains and prints statistics about the given file (either local or remote)
    """)

    parser.add_argument('file', help='File to obtain statistics for')

    opts = vars(parser.parse_args())

    s = wandio_stat(opts["file"])
    print("mtime: %s" % s["mtime"])
    print("size: %s" % s["size"])



import bz2
import zlib
import wandio.file


class CompressedReader(wandio.file.GenericReader):

    BUFLEN = 4096

    def __init__(self, child_reader, flush_dc):
        self.child_reader = child_reader
        self.flush_dc = flush_dc
        self.dc = None
        self.c_buf = b""
        self.buf = b""
        self.eof = False
        self._refill()
        super(CompressedReader, self).__init__(child_reader)

    def _get_dc(self):
        raise NotImplementedError

    def _refill(self):
        if self.eof:
            return
        # keep reading from the child reader until we've decompressed
        # enough data to fill our buffer
        while len(self.c_buf) or len(self.buf) < self.BUFLEN:
            if not len(self.c_buf):
                # fill our buffer with compressed data
                self.c_buf = self.child_reader.read(self.BUFLEN)
                if len(self.c_buf) < self.BUFLEN:
                    self.eof = True
            # pass the compressed data to the decompressor, and get back
            # some decompressed data
            if self.dc is None:
                self.dc = self._get_dc()
            # feed in the data in c_buf to decompress it
            self.buf += self.dc.decompress(self.c_buf)
            # if we are EOF on the child reader, then flush the decompressor
            if self.eof and self.flush_dc:
                self.buf += self.dc.flush()
            self.c_buf = self.dc.unused_data
            if len(self.c_buf):
                # we have leftover data after compression ended,
                # so now we need a new decompressor
                if self.flush_dc:
                    self.buf += self.dc.flush()
                self.dc = None
            # only heed the eof once we've emptied c_buf
            if not len(self.c_buf) and self.eof:
                break

    def read(self, size=None):
        res = b""
        while ((not self.eof) or len(self.buf)) and (size is None or len(res) < size):
            if not len(self.buf):
                self._refill()
            toread = size-len(res) if size is not None else len(self.buf)
            res += self.buf[0:toread]
            self.buf = self.buf[toread:]
        # TODO: remove these asserts
        assert(size is None or len(res) <= size)
        assert(size is None or len(res) == size or len(self.buf) == 0)
        return res

    def __next__(self):
        line = self.readline()
        if not line:
            assert(not len(self.buf) and self.eof)
            raise StopIteration
        return line

    def readline(self):
        res = b""
        while not len(res) or not res.endswith(b"\n"):
            idx = self.buf.find(b"\n")
            if idx == -1:
                res += self.buf
                self.buf = b""
                self._refill()
            else:
                res += self.buf[0:idx+1]
                self.buf = self.buf[idx+1:]
            if not len(self.buf) and self.eof:
                break
        if not len(res) and not len(self.buf) and self.eof:
            return None
        return res.decode("utf-8")


class CompressedWriter(wandio.file.GenericWriter):

    def __init__(self, compressor, child_writer):
        self.compressor = compressor
        self.child_writer = child_writer
        super(CompressedWriter, self).__init__(child_writer)

    def flush(self):
        cd = self.compressor.flush()
        self.fh.write(cd)
        self.fh.flush()

    def write(self, data):
        cd = self.compressor.compress(data)
        # cd is partial compressed data
        self.fh.write(cd)

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def close(self):
        self.flush()
        self.fh.close()


class GzipReader(CompressedReader):

    def __init__(self, child):
        super(GzipReader, self).__init__(child, flush_dc=True)

    def _get_dc(self):
        return zlib.decompressobj(16 + zlib.MAX_WBITS)


class GzipWriter(CompressedWriter):

    def __init__(self, child):
        compressor = zlib.compressobj(-1, zlib.DEFLATED, 16+zlib.MAX_WBITS)
        super(GzipWriter, self).__init__(compressor, child)


class BzipReader(CompressedReader):

    def __init__(self, child):
        super(BzipReader, self).__init__(child, flush_dc=False)

    def _get_dc(self):
        return bz2.BZ2Decompressor()


class BzipWriter(CompressedWriter):

    def __init__(self, child):
        compressor = bz2.BZ2Compressor()
        super(BzipWriter, self).__init__(compressor, child)



import sys

import email
import wandio.file

# urllib import compatible with both python2 and python3
# https://python-future.org/compatible_idioms.html#urllib-module
try:
    from urllib.request import urlopen, Request
except ImportError:
    from urllib2 import urlopen, Request


def http_stat(filename):
    request = Request(filename)
    request.get_method = lambda: 'HEAD'
    response = urlopen(request)
    hdrs = response.info()

    # Last Modified time
    mtime = None
    if "Last-Modified" in hdrs:
        mtime = hdrs["Last-Modified"]
        mtime = email.utils.parsedate_tz(mtime)
        mtime = email.utils.mktime_tz(mtime)

    # Content Length
    size = None
    if "Content-Length" in hdrs:
        size = int(hdrs["Content-Length"])

    return {
        "mtime": mtime,
        "size": size,
    }


class HttpReader(wandio.file.GenericReader):

    def __init__(self, url):
        self.url = url
        super(HttpReader, self).__init__(urlopen(self.url))

    def __next__(self):
        return next(self.fh).decode("utf-8")



#!/usr/bin/env python

import setuptools

setuptools.setup(
    name='pywandio',
    version='0.2.2',
    description='High-level file IO library',
    url='https://github.com/CAIDA/pywandio',
    author='Alistair King, Chiara Orsini, Mingwei Zhang',
    author_email='software@caida.org',
    packages=setuptools.find_packages(),
    install_requires=[
        'python-keystoneclient',
        'python-swiftclient',
    ],
    entry_points={'console_scripts': [
        'pywandio-cat = wandio.opener:read_main',
        'pywandio-write = wandio.opener:write_main',
        'pywandio-stat = wandio.opener:stat_main'
    ]}
)
